/*
  Ideally, you won't need to change this file.  You may want to change
  a few settings to speed debugging runs, but remember to change back
  to the original settings during final testing.

  The output format: "Size: %u\tmflop/s: %g\n"
*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <float.h>
#include <math.h>

#include <sys/types.h>
#include <sys/resource.h>

#include <unistd.h>
#include <time.h>

/*
  I attempt to run enough iterations to get reasonable timings.  The
  matrices are multiplied until either N_TIMING_MFLOPS floating-point
  operations or N_TIMING_RUNS_MAX matmuls have occured.  No fewer than
  N_TIMING_RUNS_MIN matmuls are performed.

  You may want to modify these to speed debugging...
*/
#define N_TIMING_MFLOPS 300
#define N_TIMING_RUNS_MIN 5
#define N_TIMING_RUNS_MAX 100

/*
  Your function _MUST_ have the following signature:
*/
extern void square_dgemm (const unsigned M, 
                          const double *A, const double *B, double *C);

/*
  Note the strange sizes...  You'll see some interesting effects
  around some of the powers-of-two.
*/
#if 1
/*
  Change that to zero for an abbreviated list while debugging...
*/
const unsigned test_sizes[] = {
     20,
     24,
     31,
     32,
     48,
     64,
     73,
     96,
     97,
     127,
     128,
     129,
     163,
     191,
     192,
     229,
     255,
     256,
     257,
     319,
     320,
     321,
     417,
     479,
     480,
     511,
     512,
};

# define MAX_SIZE 512u
#else
const unsigned test_sizes[] = {
     20,
     24,
     31,
     32,
     48,
     127,
     128,
     129,
     255,
     256,
     257,
     319,
     320,
     321,
     511,
     512,
};
#  define MAX_SIZE 512u
#endif


#define N_SIZES ((unsigned) sizeof (test_sizes) / sizeof (unsigned))

double A[MAX_SIZE * MAX_SIZE];
double B[MAX_SIZE * MAX_SIZE];
double C[MAX_SIZE * MAX_SIZE];

void 
matrix_init (double *A)
{
     unsigned i;

     for (i = 0; i < MAX_SIZE*MAX_SIZE; ++i) {
          A[i] = drand48 ();
     }
}

void 
matrix_clear (double *C) 
{
     memset (C, 0, MAX_SIZE * MAX_SIZE * sizeof (double));
}

/*
  Yes, another matmul.  This one is just to check that the answer
  is correct.
*/
void
baseline_dgemm (const unsigned M,
                const double alpha,
                const double* A, const double* B, double* C)
{
     unsigned i, j, k;
  
     for (i = 0; i < M; ++i) {
          for (j = 0; j < M; ++j) {
               for (k = 0; k < M; ++k) {
                    C[j*M + i] += alpha * A[k*M + i] * B[j*M + k];
               }
          }
     }
}

int
validate_dgemm (const unsigned M,
                const double *A, const double *B, double *C)
{
     unsigned i = 0;
     double A_norm = 0.0;
     double B_norm = 0.0;
     double C_norm = 0.0;

     matrix_clear (C);
     baseline_dgemm (M, -1.0, A, B, C);
     square_dgemm (M, A, B, C);

     /*
       The 1-norm is simply the sum of absolute values.  Nice and easy.
     */
     for (i = 0; i < M*M; ++i) {
          A_norm += fabs(A[i]);
          B_norm += fabs(B[i]);
          C_norm += fabs(C[i]);
     }

     if (C_norm > DBL_EPSILON * A_norm * B_norm) {
          return 0;
     }

     return 1;
}

int
main (void)
{
     unsigned sz_i, i;
     /*
       clock() normally measures milliseconds.  In non-Alpha Linux,
       though, it's limited by the HZ in include/asm-i386/param.h.
       The setting is, imho, unreasonably low for faster machines,
       but...  It's a subject of flame wars.  sigh.

       Timing under Linux is not fun.
     */
     clock_t cpu_time;
     clock_t last_clock;
     double mflops;
     double sec, mflop_s;

     matrix_init (A);
     matrix_init (B);

     for (sz_i = 0; sz_i < N_SIZES; ++sz_i) {

          const unsigned M = test_sizes[sz_i];

          /* Test first... */

          if (!validate_dgemm (M, A, B, C)) {
               fprintf (stderr, "Matrix multiplication failed.\n");
               return -1;
          }

          /* ... and now time. */

          cpu_time = 0;
          mflops = 0.0;
          i = 0;
          while (mflops < N_TIMING_MFLOPS || i < N_TIMING_RUNS_MIN) {
               if (i > N_TIMING_RUNS_MAX)
                    break;
               
               matrix_clear (C);

               last_clock = clock();
      
               square_dgemm (M, A, B, C);

               cpu_time += clock() - last_clock;
               mflops += 2.0 * M * M * M / 1.0e6;
               ++i;
          }

          sec = cpu_time / ((double) CLOCKS_PER_SEC);
          sec /= (double)i;

          mflop_s = 2.0 * M*M*M / sec / 1.0e6;

          printf ("Size: %u\tmflop/s: %g\n", 
                  M, mflop_s);
     }

     return 0;
}
